"""Makes a login request.

This code gets the username and password of a user and makes a login request to the token endpoint.
If an OTP token is required, the user must provide either an OTP code from their authenticator app or
a backed up static code in order to complete the authentication process.

    Typical usage example:

    login_request = login.UsernamePasswordLoginAPIRequest(username, password, otp_token)
"""

from typing import Dict, Optional

from . import request


class UsernamePasswordLoginAPIRequest(request.APIRequest):
    """Makes a request to log in the user."""

    def __init__(self, username: str, password: str, otp_token: Optional[str] = None) -> None:
        """Constructs all the necessary attributes for the object.

        Args:
            username: the username (email) used to login.
            password: the password used to login.
            otp_token: the OTP or static code if the user has enabled OTP. Defaults to None.
        """
        self._username = username
        self._password = password
        self._otp_token = otp_token

    @property
    def endpoint(self):
        return request.TOKEN_ENDPOINT

    @property
    def query(self) -> Optional[str]:
        return None

    @property
    def data(self) -> Optional[Dict]:
        """Gets the user login credentials.

        Returns:
            The user login credentials.
        """
        if self._otp_token is not None:
            data = {
                'username': self._username,
                'password': self._password,
                'otp_token': self._otp_token,
            }
            return data
        else:
            data = {
                'username': self._username,
                'password': self._password,
            }
            return data



"""Runtime are in charge of running scan as defines by a set of agents, agent group and a target asset."""
import abc

from ostorlab.runtimes import definitions
from ostorlab.assets import asset as base_asset



class Runtime(abc.ABC):
    """Runtime is in charge of preparing the environment to trigger a scan."""

    @abc.abstractmethod
    def can_run(self, agent_group_definition: definitions.AgentGroupDefinition) -> bool:
        """Checks if the runtime is capable of running the provided agent run definition.

        Args:
            agent_group_definition: The agent run definition from a set of agents and agent groups.

        Returns:
            True if can run, false otherwise.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def scan(self, agent_group_definition: definitions.AgentGroupDefinition, asset: base_asset.Asset) -> None:
        """Triggers a scan using the provided agent run definition and asset target.

        Args:
            agent_group_definition: The agent run definition from a set of agents and agent groups.
            asset: The scan target asset.

        Returns:
            None
        """
        raise NotImplementedError()



"""This module is the entry point for ostorlab CLI."""
import click


@click.group()
@click.pass_context
def rootcli(ctx: click.core.Context) -> None:
    """Ostorlab is an open-source project to help automate security testing.\n
    Ostorlab standardizes interoperability between tools in a consistent, scalable, and performant way."""

    ctx.obj = {}


@rootcli.group()
def agent():
    """ToDo"""
    pass

@rootcli.group()
def agentgroup():
    raise click.ClickException('NotImplementedError.')




# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v2/report/status.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='v2/report/status.proto',
    package='',
    serialized_pb=_b(
        '\n\x16v2/report/status.proto\"i\n\x06status\x12\x0f\n\x07scan_id\x18\x01 \x02(\x05\x12%\n\nattributes\x18\x02 \x03(\x0b\x32\x11.status.attribute\x1a\'\n\tattribute\x12\x0b\n\x03key\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x02(\t')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_STATUS_ATTRIBUTE = _descriptor.Descriptor(
    name='attribute',
    full_name='status.attribute',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='key', full_name='status.attribute.key', index=0,
            number=1, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='value', full_name='status.attribute.value', index=1,
            number=2, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=92,
    serialized_end=131,
)

_STATUS = _descriptor.Descriptor(
    name='status',
    full_name='status',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='scan_id', full_name='status.scan_id', index=0,
            number=1, type=5, cpp_type=1, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='attributes', full_name='status.attributes', index=1,
            number=2, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[_STATUS_ATTRIBUTE, ],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=26,
    serialized_end=131,
)

_STATUS_ATTRIBUTE.containing_type = _STATUS
_STATUS.fields_by_name['attributes'].message_type = _STATUS_ATTRIBUTE
DESCRIPTOR.message_types_by_name['status'] = _STATUS

status = _reflection.GeneratedProtocolMessageType('status', (_message.Message,), dict(

    attribute=_reflection.GeneratedProtocolMessageType('attribute', (_message.Message,), dict(
        DESCRIPTOR=_STATUS_ATTRIBUTE,
        __module__='v2.report.status_pb2'
        # @@protoc_insertion_point(class_scope:status.attribute)
    ))
    ,
    DESCRIPTOR=_STATUS,
    __module__='v2.report.status_pb2'
    # @@protoc_insertion_point(class_scope:status)
))
_sym_db.RegisterMessage(status)
_sym_db.RegisterMessage(status.attribute)

# @@protoc_insertion_point(module_scope)



"""Agent Build commands."""
import io
import logging

import click
import docker

from ostorlab.agent.schema import loader
from ostorlab.agent.schema import validator
from ostorlab.cli.agent import agent

logger = logging.getLogger(__name__)


@agent.command()
@click.option('--file', '-f', type=click.File('rb'), help='Path to Agent yaml definition.', required=True)
def build(file: io.FileIO) -> None:
    """CLI command to build the agent container from a definition.yaml file.
    Usage : Ostorlab agent build -f path/to/definition.yaml
    """
    try:
        agent_def = loader.load_agent_yaml(file)

        dockerfile_path = agent_def['docker_file_path']
        docker_build_root = agent_def['docker_build_root']
        container_name = agent_def['name']

        docker_sdk_client = docker.from_env()
        docker_sdk_client.images.build(path=docker_build_root, dockerfile=dockerfile_path, tag=container_name)
    except validator.SchemaError:
        logger.error(
            'Schema is invalid, this should not happen, please report an issue at '
            'https://github.com/Ostorlab/ostorlab/issues.')
    except validator.ValidationError:
        logger.error('Definition file does not conform to the provided specification.')



"""Agent and Agent group definitions and settings dataclasses."""
import dataclasses
import io
from typing import List, Optional

from ostorlab.agent.schema import loader
from ostorlab.runtimes.proto import agent_instance_settings_pb2
from ostorlab.utils import defintions


@dataclasses.dataclass
class AgentSettings:
    """Agent instance lists the settings of running instance of an agent."""
    key: str
    bus_url: Optional[str] = None
    bus_exchange_topic: Optional[str] = None
    args: List[defintions.Arg] = dataclasses.field(default_factory=list)
    constraints: List[str] = dataclasses.field(default_factory=list)
    mounts: Optional[List[str]] = dataclasses.field(default_factory=list)
    restart_policy: str = 'any'
    mem_limit: Optional[int] = None
    open_ports: List[defintions.PortMapping] = dataclasses.field(default_factory=list)
    replicas: int = 1
    healthcheck_host: str = '0.0.0.0'
    healthcheck_port: int = 5000

    @property
    def container_image(self):
        """Agent image name."""
        image = self.key.replace('/', '_')
        # TODO (alaeddine): add container tag resolution.
        return image

    @classmethod
    def from_proto(cls, proto: bytes) -> 'AgentSettings':
        """Constructs an agent definition from a binary proto settings.

        Args:
            proto: binary proto settings file.

        Returns:
            AgentInstanceSettings object.
        """
        instance = agent_instance_settings_pb2.AgentInstanceSettings()
        instance.ParseFromString(proto)
        return cls(
            key=instance.key,
            bus_url=instance.bus_url,
            bus_exchange_topic=instance.bus_exchange_topic,
            args=[defintions.Arg(
                name=a.name,
                type=a.type,
                value=a.value
            ) for a in instance.args],
            constraints=instance.constraints,
            mounts=instance.mounts,
            restart_policy=instance.restart_policy,
            mem_limit=instance.mem_limit,
            open_ports=[defintions.PortMapping(
                source_port=p.source_port,
                destination_port=p.destination_port
            ) for p in instance.open_ports],
            replicas=instance.replicas,
            healthcheck_host=instance.healthcheck_host,
            healthcheck_port=instance.healthcheck_port,
        )

    def to_raw_proto(self) -> bytes:
        """Transforms agent instance settings into a raw proto bytes.

        Returns:
            Bytes as a serialized proto.
        """
        instance = agent_instance_settings_pb2.AgentInstanceSettings()
        instance.key = self.key
        instance.bus_url = self.bus_url
        instance.bus_exchange_topic = self.bus_exchange_topic

        for arg in self.args:
            arg_instance = instance.args.add()
            arg_instance.name = arg.name
            arg_instance.type = arg.type
            arg_instance.value = arg.value

        instance.constraints.extend(self.constraints)
        instance.mounts.extend(self.mounts)
        instance.restart_policy = self.restart_policy
        if self.mem_limit is not None:
            instance.mem_limit = self.mem_limit

        for open_port in self.open_ports:
            open_port_instance = instance.open_ports.add()
            open_port_instance.source_port = open_port.source_port
            open_port_instance.destination_port = open_port.destination_port

        instance.replicas = self.replicas
        instance.healthcheck_host = self.healthcheck_host
        instance.healthcheck_port = self.healthcheck_port

        return instance.SerializeToString()


@dataclasses.dataclass
class AgentGroupDefinition:
    """Data class holding the attributes of an agent."""
    agents: List[AgentSettings]

    @classmethod
    def from_yaml(cls, group: io.FileIO):
        """Construct AgentGroupDefinition from yaml file.

        Args:
            group : agent group .yaml file.
        """
        agent_group_def = loader.load_agent_group_yaml(group)
        agent_settings = []
        for agent in agent_group_def['agents']:
            agent_def = AgentSettings(
                key=agent.get('key'),
                args=[defintions.Arg(name=a.get('name'), description=a.get('description'), type=a.get('type'),
                                     value=a.get('value')) for a in
                      agent.get('args', [])],
                constraints=agent.get('constraints'),
                mounts=agent.get('mounts'),
                restart_policy=agent.get('restart_policy'),
                mem_limit=agent.get('mem_limit'),
                open_ports=[defintions.PortMapping(source_port=p.get('src_port'), destination_port=p.get('dest_port'))
                            for p
                            in agent.get('open_ports', [])]
            )

            agent_settings.append(agent_def)

        return cls(agent_settings)



"""Android .AAB asset."""
import dataclasses
from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.file.android.aab')
class AndroidAab(asset.Asset):
    """Android .AAB target asset."""
    content: bytes



"""Serializer handles matching of selector to the proper protobuf message definition."""
import importlib
import os
import pathlib
import re
from typing import Dict, Optional, List, Any

from google.protobuf import json_format

from ostorlab import exceptions

PROTO_CLASS_NAME = 'Message'

# When the package is built, the proto files are not shipped, only the compiled version, which have the
# form: logs_pb2.cpython-38.pyc
MESSAGE_PROTO = '_pb2'
MESSAGE_CODE_PATH = pathlib.Path(__file__).parent / 'proto'


class SerializationError(exceptions.OstorlabError):
    """Base serialization Error."""


class TooManyMatchingPackageNamesError(SerializationError):
    """There are over 2 matching package names."""


class NoMatchingPackageNameError(SerializationError):
    """There are no matching package name."""


def _find_package_name(selector: str) -> Optional[str]:
    """Finds matching package name to selector."""
    files = _list_message_proto_files()
    pattern = re.compile(_selector_to_package_regex(selector))
    matching = [pattern.match(f) for f in files]
    filtered_matching = [m.group(0) for m in matching if m is not None]
    if len(filtered_matching) > 1:
        raise TooManyMatchingPackageNamesError()
    elif len(filtered_matching) == 0:
        raise NoMatchingPackageNameError()
    else:
        # Remove the path to the current package.
        matching_package = re.sub(r'^.*/ostorlab/', 'ostorlab/', filtered_matching[0])
        # Replace / with .
        matching_package = matching_package.replace('/', '.')
        # Point to the compiled protobufs.
        return re.sub(r'_pb2\..*$', '_pb2', re.sub(r'^\.code\.', '', matching_package))


def _list_message_proto_files() -> List[str]:
    """List all the proto files."""
    files = []
    # r=root, d=directories, f = files
    for r, d, f in os.walk(MESSAGE_CODE_PATH):
        for file in f:
            if MESSAGE_PROTO in file:
                files.append(os.path.join(r, file))
        del d
    return files


def _selector_to_package_regex(subject):
    """Maps selector to package matching regular expression."""
    splitted = subject.split('.')
    return '.*/message/proto/' + '/'.join([f'(_[_a-zA-Z0-9]+|{s})' for s in splitted]) + r'.[_a-zA-Z0-9]+\_pb2\..*'


def serialize(selector: str, values: Dict[str, Any]):
    """Serializes a Request message using the proper format defined using the seelctor value.
    If the subject is a.b.c. The corresponding proto is located at message/a/b/c/xxx.proto.

    Args:
        selector: Message selector, must specify the version in use.
        values: Dict representation of the message to serialize.

    Returns:
        Proto serialized message.
    """
    try:
        return _serialize(selector, PROTO_CLASS_NAME, values)
    except json_format.Error as e:
        raise SerializationError('Error serializing message') from e


def _serialize(selector: str, class_name: str, values: Dict[str, Any]):
    """Serializes message using the selector and defined class name."""
    package_name = _find_package_name(selector)
    class_object = getattr(importlib.import_module(package_name), class_name)
    proto_message = class_object()
    _parse_dict(values, proto_message)
    return proto_message


def _parse_list(values: Any, message) -> None:
    """Parse list to protobuf message."""
    if isinstance(values[0], dict):  # value needs to be further parsed
        for v in values:
            cmd = message.add()
            _parse_dict(v, cmd)
    else:  # value can be set
        message.extend(values)


def _parse_dict(values: Any, message) -> None:
    """Parse dict to protobuf message."""
    for k, v in values.items():
        if isinstance(v, dict):  # value needs to be further parsed
            _parse_dict(v, getattr(message, k))
        elif isinstance(v, list):
            _parse_list(v, getattr(message, k))
        else:
            try:
                setattr(message, k, v)
            except AttributeError as e:
                raise SerializationError('invalid attribute {k}') from e


def deserialize(selector: str, serialized: bytes):
    """Deserializes a Request message using the proper format defined using the selector value.

    Args:
        selector: Message selector, must specify the version in use.
        serialized: Raw message to deserialize.

    Returns:
        Dict
    """
    return _deserialize(selector, PROTO_CLASS_NAME, serialized)


def _deserialize(selector: str, class_name: str, serialized: bytes):
    """Deserializes message using the selector and defined class name."""
    package_name = _find_package_name(selector)
    class_object = getattr(importlib.import_module(package_name), class_name)
    proto_message = class_object()
    proto_message.ParseFromString(serialized)
    return proto_message



"""Scan module that handles running a scan using a list of agent keys and a target asset."""
from typing import Optional

from ostorlab.cli.rootcli import rootcli
import click


@rootcli.group()
@click.option('--proxy', '-X', help='Proxy to route HTTPS requests through.')
@click.option('--tlsverify/--no-tlsverify', help='Control TLS server certificate verification.', default=True)
@click.pass_context
def scan(ctx: click.core.Context, proxy: Optional[str], tlsverify: bool) -> None:
    """You can use scan [subcommand] to list, start or stop a scan.\n
    Examples:\n
        - Show list of scans: ostorlab scan --list\n
        - Show full details of a scan: ostorlab scan describe --scan=scan_id.\n
    """
    ctx.obj['proxy'] = proxy
    ctx.obj['tlsverify'] = tlsverify
    pass



"""Asset of type .AAB package file.
This module takes care of preparing a file of type .aab before injecting it to the runtime instance."""
import io
import click
from ostorlab.assets import android_aab as android_aab_asset
from ostorlab.cli.scan.run import run


@run.run.command()
@click.option('--file', type=click.File(), help='Path to .AAB file.', required=True)
@click.pass_context
def android_aab(ctx: click.core.Context, file: io.FileIO) -> None:
    """Run scan for android .AAB package file."""
    runtime = ctx.obj['runtime']
    asset = android_aab_asset.AndroidAab(file)
    runtime.scan(agent_group_definition=ctx.obj['agent_run_definition'], asset=asset)



"""Module for the command run inside the group scan.
This module takes care of preparing the selected runtime and the lists of provided agents, before starting a scan.
Example of usage:
    - ostorlab scan run --agents=agent1,agent2 --title=test_scan [asset] [options]."""
import io
from typing import List
import click
from ostorlab import runtimes
from ostorlab.runtimes import definitions
from ostorlab.cli.scan import scan


@scan.group()
@click.option('--runtime', type=click.Choice(['local', 'managed', 'hybrid']),
              help="""Runtime is in charge of preparing the environment to trigger a scan.\n
                    Specify which runtime to use: \n
                    local: on you local machine \n
                    managed: on Ostorlab cloud, (requires login) \n
                    hybrid: soon!. \n
                   """,
              required=True)
@click.option('--agents', multiple=True, help='List of agents keys. to use in the scan. ', required=True)
@click.option('--title', '-t', help='Scan title.')
@click.option('--agents-group-definition', '-g', type=click.File('r'),
              help='Path to agents group definition file (yaml).', multiple=True)
@click.pass_context
def run(ctx: click.core.Context, runtime: str, agents: List[str], agents_group_definition: List[io.FileIO],
        title: str) -> None:
    """Start a new scan on a specific asset.\n
    Usage:\n
        - ostorlab scan run --agents=agent1,agent2 --title=test_scan android-apk --file=path/to/application.apk
    """

    if runtime == 'local':
        runtime = runtimes.LocalRuntime()
    else:
        # managed and hybrid are not implemented yet
        raise click.ClickException(f'The selected runtime {runtime} is not supported!')

    # Building list of agents definition
    agents_settings: List[definitions.AgentSettings] = []
    for agent_key in agents:
        agents_settings.append(
            definitions.AgentSettings(key=agent_key))

    # Building list of agent group definition
    agents_groups: List[definitions.AgentGroupDefinition] = []
    for group in agents_group_definition:
        agents_groups.append(definitions.AgentGroupDefinition.from_file(group))

    agent_run_definition = definitions.AgentRunDefinition(
        agent_groups=agents_groups, agents=agents_settings)
    if runtime.can_run(agent_group_definition=agent_run_definition):
        ctx.obj['runtime'] = runtime
        ctx.obj['agent_run_definition'] = agent_run_definition
        ctx.obj['title'] = title
    else:
        raise click.ClickException('Error: invalid agent list.')



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: test.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='test.proto',
    package='',
    syntax='proto2',
    serialized_pb=_b(
        '\n\ntest.proto\"5\n\x16link_test_extra_header\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x02(\t\"/\n\x10link_test_cookie\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x02(\t\"A\n\x14link_test_form_input\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x02(\t\x12\x0c\n\x04type\x18\x03 \x02(\t\"H\n\x0elink_test_form\x12%\n\x06inputs\x18\x01 \x03(\x0b\x32\x15.link_test_form_input\x12\x0f\n\x07\x65nctype\x18\x02 \x02(\t\"M\n\x0ftest_credential\x12\x0c\n\x04type\x18\x01 \x02(\t\x12\r\n\x05login\x18\x02 \x01(\t\x12\x10\n\x08password\x18\x03 \x01(\t\x12\x0b\n\x03url\x18\x04 \x01(\t\"\x95\x02\n\x04test\x12\x0f\n\x07scan_id\x18\x01 \x02(\x05\x12\x0f\n\x07link_id\x18\x02 \x01(\x05\x12\x0b\n\x03url\x18\x03 \x02(\t\x12\x0e\n\x06method\x18\x04 \x02(\t\x12\x0e\n\x04\x62ody\x18\x05 \x01(\x0cH\x00\x12\x1f\n\x04\x66orm\x18\x06 \x01(\x0b\x32\x0f.link_test_formH\x00\x12.\n\rextra_headers\x18\x07 \x03(\x0b\x32\x17.link_test_extra_header\x12\"\n\x07\x63ookies\x18\x08 \x03(\x0b\x32\x11.link_test_cookie\x12$\n\ncredential\x18\t \x01(\x0b\x32\x10.test_credential\x12\x15\n\x06parent\x18\n \x01(\x0b\x32\x05.testB\x0c\n\nbody_oneof')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_LINK_TEST_EXTRA_HEADER = _descriptor.Descriptor(
    name='link_test_extra_header',
    full_name='link_test_extra_header',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='name', full_name='link_test_extra_header.name', index=0,
            number=1, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='value', full_name='link_test_extra_header.value', index=1,
            number=2, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=14,
    serialized_end=67,
)

_LINK_TEST_COOKIE = _descriptor.Descriptor(
    name='link_test_cookie',
    full_name='link_test_cookie',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='name', full_name='link_test_cookie.name', index=0,
            number=1, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='value', full_name='link_test_cookie.value', index=1,
            number=2, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=69,
    serialized_end=116,
)

_LINK_TEST_FORM_INPUT = _descriptor.Descriptor(
    name='link_test_form_input',
    full_name='link_test_form_input',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='name', full_name='link_test_form_input.name', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='value', full_name='link_test_form_input.value', index=1,
            number=2, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='type', full_name='link_test_form_input.type', index=2,
            number=3, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=118,
    serialized_end=183,
)

_LINK_TEST_FORM = _descriptor.Descriptor(
    name='link_test_form',
    full_name='link_test_form',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='inputs', full_name='link_test_form.inputs', index=0,
            number=1, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='enctype', full_name='link_test_form.enctype', index=1,
            number=2, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=185,
    serialized_end=257,
)

_TEST_CREDENTIAL = _descriptor.Descriptor(
    name='test_credential',
    full_name='test_credential',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='type', full_name='test_credential.type', index=0,
            number=1, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='login', full_name='test_credential.login', index=1,
            number=2, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='password', full_name='test_credential.password', index=2,
            number=3, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='url', full_name='test_credential.url', index=3,
            number=4, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=259,
    serialized_end=336,
)

_TEST = _descriptor.Descriptor(
    name='test',
    full_name='test',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='scan_id', full_name='test.scan_id', index=0,
            number=1, type=5, cpp_type=1, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='link_id', full_name='test.link_id', index=1,
            number=2, type=5, cpp_type=1, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='url', full_name='test.url', index=2,
            number=3, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='method', full_name='test.method', index=3,
            number=4, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='body', full_name='test.body', index=4,
            number=5, type=12, cpp_type=9, label=1,
            has_default_value=False, default_value=_b(""),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='form', full_name='test.form', index=5,
            number=6, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='extra_headers', full_name='test.extra_headers', index=6,
            number=7, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='cookies', full_name='test.cookies', index=7,
            number=8, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='credential', full_name='test.credential', index=8,
            number=9, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='parent', full_name='test.parent', index=9,
            number=10, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name='body_oneof', full_name='test.body_oneof',
            index=0, containing_type=None, fields=[]),
    ],
    serialized_start=339,
    serialized_end=616,
)

_LINK_TEST_FORM.fields_by_name['inputs'].message_type = _LINK_TEST_FORM_INPUT
_TEST.fields_by_name['form'].message_type = _LINK_TEST_FORM
_TEST.fields_by_name['extra_headers'].message_type = _LINK_TEST_EXTRA_HEADER
_TEST.fields_by_name['cookies'].message_type = _LINK_TEST_COOKIE
_TEST.fields_by_name['credential'].message_type = _TEST_CREDENTIAL
_TEST.fields_by_name['parent'].message_type = _TEST
_TEST.oneofs_by_name['body_oneof'].fields.append(
    _TEST.fields_by_name['body'])
_TEST.fields_by_name['body'].containing_oneof = _TEST.oneofs_by_name['body_oneof']
_TEST.oneofs_by_name['body_oneof'].fields.append(
    _TEST.fields_by_name['form'])
_TEST.fields_by_name['form'].containing_oneof = _TEST.oneofs_by_name['body_oneof']
DESCRIPTOR.message_types_by_name['link_test_extra_header'] = _LINK_TEST_EXTRA_HEADER
DESCRIPTOR.message_types_by_name['link_test_cookie'] = _LINK_TEST_COOKIE
DESCRIPTOR.message_types_by_name['link_test_form_input'] = _LINK_TEST_FORM_INPUT
DESCRIPTOR.message_types_by_name['link_test_form'] = _LINK_TEST_FORM
DESCRIPTOR.message_types_by_name['test_credential'] = _TEST_CREDENTIAL
DESCRIPTOR.message_types_by_name['test'] = _TEST

link_test_extra_header = _reflection.GeneratedProtocolMessageType('link_test_extra_header', (_message.Message,), dict(
    DESCRIPTOR=_LINK_TEST_EXTRA_HEADER,
    __module__='test_pb2'
    # @@protoc_insertion_point(class_scope:link_test_extra_header)
))
_sym_db.RegisterMessage(link_test_extra_header)

link_test_cookie = _reflection.GeneratedProtocolMessageType('link_test_cookie', (_message.Message,), dict(
    DESCRIPTOR=_LINK_TEST_COOKIE,
    __module__='test_pb2'
    # @@protoc_insertion_point(class_scope:link_test_cookie)
))
_sym_db.RegisterMessage(link_test_cookie)

link_test_form_input = _reflection.GeneratedProtocolMessageType('link_test_form_input', (_message.Message,), dict(
    DESCRIPTOR=_LINK_TEST_FORM_INPUT,
    __module__='test_pb2'
    # @@protoc_insertion_point(class_scope:link_test_form_input)
))
_sym_db.RegisterMessage(link_test_form_input)

link_test_form = _reflection.GeneratedProtocolMessageType('link_test_form', (_message.Message,), dict(
    DESCRIPTOR=_LINK_TEST_FORM,
    __module__='test_pb2'
    # @@protoc_insertion_point(class_scope:link_test_form)
))
_sym_db.RegisterMessage(link_test_form)

test_credential = _reflection.GeneratedProtocolMessageType('test_credential', (_message.Message,), dict(
    DESCRIPTOR=_TEST_CREDENTIAL,
    __module__='test_pb2'
    # @@protoc_insertion_point(class_scope:test_credential)
))
_sym_db.RegisterMessage(test_credential)

test = _reflection.GeneratedProtocolMessageType('test', (_message.Message,), dict(
    DESCRIPTOR=_TEST,
    __module__='test_pb2'
    # @@protoc_insertion_point(class_scope:test)
))
_sym_db.RegisterMessage(test)

# @@protoc_insertion_point(module_scope)



"""Asset of type .IPA (IOS App Store Package).
This module takes care of preparing a file of type .IPA before injecting it to the runtime instance.
"""

import click
from ostorlab.assets import ios_ipa as ios_ipa_asset
from ostorlab.cli.scan.run import run


@run.run.command()
@click.option('--file', type=click.File(), help='Path to .IPA file.', required=True)
@click.pass_context
def ios_ipa(ctx, file):
    """Run scan for .IPA package file."""

    runtime = ctx.obj['runtime']
    asset = ios_ipa_asset.IOSIpa(file.read())
    runtime.scan(agent_group_definition=ctx.obj['agent_run_definition'], asset=asset)



"""Common definition types."""
import dataclasses
from typing import Optional


@dataclasses.dataclass
class Arg:
    """Data class holding a definition.

    The value is always bytes to support all arg values. The type is defined by the type attribute."""
    name: str
    type: str
    value: bytes
    description: Optional[str] = None


@dataclasses.dataclass
class PortMapping:
    """Data class defining a port mapping source to destination"""
    source_port: int
    destination_port: int



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v3/capture/logs/logs.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='v3/capture/logs/logs.proto',
  package='v3.capture.logs',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x1av3/capture/logs/logs.proto\x12\x0fv3.capture.logs\"9\n\x04logs\x12\x11\n\ttimestamp\x18\x01 \x02(\x02\x12\r\n\x05level\x18\x02 \x02(\t\x12\x0f\n\x07message\x18\x03 \x02(\t')
)




_LOGS = _descriptor.Descriptor(
  name='logs',
  full_name='v3.capture.logs.logs',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='v3.capture.logs.logs.timestamp', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='level', full_name='v3.capture.logs.logs.level', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='v3.capture.logs.logs.message', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=47,
  serialized_end=104,
)

DESCRIPTOR.message_types_by_name['logs'] = _LOGS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

logs = _reflection.GeneratedProtocolMessageType('logs', (_message.Message,), dict(
  DESCRIPTOR = _LOGS,
  __module__ = 'v3.capture.logs.logs_pb2'
  # @@protoc_insertion_point(class_scope:v3.capture.logs.logs)
  ))
_sym_db.RegisterMessage(logs)


# @@protoc_insertion_point(module_scope)



"""Auth revoke command."""

import logging

from ostorlab.apis import runner as apis_runner
from ostorlab.apis import revoke_api_key
from ostorlab import configuration_manager
from ostorlab.cli.auth import auth

logger = logging.getLogger(__name__)


@auth.auth.command()
def revoke():
    """Use this to revoke your API key.
    """
    config_manager = configuration_manager.ConfigurationManager()

    try:
        api_key_id = config_manager.get_api_key_id()
        runner = apis_runner.APIRunner()

        response = runner.execute(revoke_api_key.RevokeAPIKeyAPIRequest(api_key_id))
        if response.get('errors') is not None:
            logger.error('Error revoking your API key.')


        runner.unauthenticate()
        config_manager.delete_api_data()
    except apis_runner.AuthenticationError:
        runner.unauthenticate()
        logger.error('Your API key has already been revoked.')



"""Asset of type .APK package file.
This module takes care of preparing a file of type .APK before injecting it to the runtime instance.
"""
import io
import click
from ostorlab.assets import android_apk as android_apk_asset
from ostorlab.cli.scan.run import run


@run.run.command()
@click.option('--file', type=click.File(), help='Path to .APK file.', required=True)
@click.pass_context
def android_apk(ctx: click.core.Context, file: io.FileIO) -> None:
    """Run scan for android .APK package file."""

    runtime = ctx.obj['runtime']
    asset = android_apk_asset.AndroidApk(file.read())
    runtime.scan(agent_group_definition=ctx.obj['agent_run_definition'], asset=asset)



"""Android .APK asset."""
import dataclasses
from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.file.android.apk')
class AndroidApk(asset.Asset):
    """Android .APK  target asset."""
    content: bytes



"""TODO(mohsine):Wrtite docstring."""
import abc
import logging
from typing import Dict, Optional

AUTHENTICATED_GRAPHQL_ENDPOINT = 'https://api.ostorlab.co/apis/graphql'
PUBLIC_GRAPHQL_ENDPOINT = 'https://api.ostorlab.co/apis/public'
TOKEN_ENDPOINT = 'https://api.ostorlab.co/apis/token/'

logger = logging.getLogger(__name__)


class APIRequest(abc.ABC):
    """TODO(mohsine):Wrtite docstring."""

    @property
    def endpoint(self):
        return AUTHENTICATED_GRAPHQL_ENDPOINT

    @property
    @abc.abstractmethod
    def query(self) -> Optional[str]:
        raise NotImplementedError('Missing implementation')

    @property
    @abc.abstractmethod
    def data(self) -> Optional[Dict]:
        raise NotImplementedError('Missing implementation')



"""Ostorlab main package."""
import logging
from rich.logging import RichHandler
from ostorlab.cli.rootcli import rootcli

FORMAT = '%(message)s'
logging.basicConfig(
    level='NOTSET', format=FORMAT, datefmt='[%X]', handlers=[RichHandler(rich_tracebacks=True)]
)

logger = logging.getLogger('CLI')


def main():
    rootcli(None)



"""Handles all API calls and behind the scenes operations such as authentication, validation, etc.

This module contains code to handle all API calls and any behind the scenes logic
like authentication. It also has classes for authentication errors, API response errors, etc.

    Typical usage example:

    foo = APIRunner(username, password, token_duration)
    foo.authenticate()
"""

import logging

from typing import Dict, Optional

import requests
import click

from ostorlab import configuration_manager as config_manager
from ostorlab.apis import create_api_key
from ostorlab.apis import login
from ostorlab.apis import request as api_request

logger = logging.getLogger(__name__)


class Error(Exception):
    """Base Error Class"""


class AuthenticationError(Error):
    """Authentication Error."""


class ResponseError(Error):
    """Response Error."""


class APIRunner:
    """Handles all API calls and behind the scenes operations such as authentication,
       validation, etc.
    """

    def __init__(self,
                 username: str = None,
                 password: str = None,
                 token_duration: str = None,
                 proxy: str = None,
                 verify: bool = True):
        """Constructs all the necessary attributes for the object.

        Args:
            username: the username (email) used to login.
            password: the password used to login.
            token_duration: The duration for which the token is valid
            (Can be in minutes, hours, days, or a combination of any two or all three).
            proxy: The proxy through which a request is made. Defaults to None.
            verify: Whether or not to verify the TLS certificate. Defaults to True.
        """

        self._username = username
        self._password = password
        self._proxy = proxy
        self._verify = verify
        self._token_duration = token_duration
        self._configuration_manager: config_manager.ConfigurationManager = config_manager.ConfigurationManager()
        self._api_key: Optional[str] = self._configuration_manager.get_api_key()
        self._token: Optional[str] = None
        self._otp_token: Optional[str] = None

    def _login_user(self) -> requests.models.Response:
        """Logs in the user.

        Returns:
            The API response.
        """
        login_request = login.UsernamePasswordLoginAPIRequest(
            self._username, self._password, self._otp_token)
        return self._sent_request(login_request)

    def authenticate(self) -> None:
        """Authenticates the user.

        Raises:
            AuthenticationError: If user credentials are not valid.
        """
        response = self._login_user()

        if response.status_code != 200:
            field_errors = response.json().get('non_field_errors')
            if field_errors is not None and field_errors[0] == 'Must include \"otp_token\"':
                self._otp_token = click.prompt(
                    'Please enter the OTP code from your authenticator app')
                self.authenticate()
            else:
                logger.debug('response %s', response.content)
                raise AuthenticationError(response.status_code)
        else:
            self._token = response.json().get('token')
            api_key_response = self.execute(
                create_api_key.CreateAPIKeyAPIRequest(self._token_duration))

            api_data = api_key_response['data']['createApiKey']['apiKey']
            secret_key = api_data['secretKey']
            api_key_id = api_data['apiKey']['id']
            expiry_date = api_data['apiKey']['expiryDate']

            self._api_key = secret_key
            self._configuration_manager.set_api_data(
                secret_key, api_key_id, expiry_date)
            self._token = None

    def unauthenticate(self) -> None:
        self._api_key = None

    def execute(self, request: api_request.APIRequest) -> Dict:
        """Executes a request using the GraphQL API.

        Args:
            request: The request to be executed

        Raises:
            ResponseError: When the API returns an error

        Returns:
            The API response
        """
        if self._token is not None:
            headers = {'Authorization': f'Token {self._token}'}
        elif self._api_key is not None:
            headers = {'X-Api-Key': f'{self._api_key}'}
        else:
            headers = None
            logger.warning('No authentication credentials were provided.')


        response = self._sent_request(request, headers)
        if response.status_code != 200:
            raise ResponseError(
                f'Response status code is {response.status_code}: {response.content}')
        else:
            return response.json()

    def _sent_request(self, request: api_request.APIRequest, headers=None,
                      multipart=False) -> requests.Response:
        """Sends an API request."""
        if self._proxy is not None:
            proxy = {
                'https': self._proxy
            }
        else:
            proxy = None
        if multipart:
            return requests.post(request.endpoint, files=request.data, headers=headers,
                                 proxies=proxy, verify=self._verify)
        else:
            return requests.post(request.endpoint, data=request.data, headers=headers,
                                 proxies=proxy, verify=self._verify)



"""Agent and Agent group definitions and settings dataclasses."""
import dataclasses
import io
from typing import List, Optional

from ostorlab.agent.schema import loader
from ostorlab.utils import defintions


@dataclasses.dataclass
class AgentDefinition:
    """Data class holding attributes of an agent."""
    name: str
    in_selectors: List[str] = dataclasses.field(default_factory=list)
    out_selectors: List[str] = dataclasses.field(default_factory=list)
    args: List[defintions.Arg] = dataclasses.field(default_factory=list)
    constraints: List[str] = None
    mounts: Optional[List[str]] = None
    restart_policy: str = 'any'
    mem_limit: int = None
    open_ports: List[defintions.PortMapping] = dataclasses.field(default_factory=list)

    @classmethod
    def from_yaml(cls, file: io.FileIO) -> 'AgentDefinition':
        """Constructs an agent definition from a yaml definition file.

        Args:
            file: Yaml file.

        Returns:
            Agent definition.
        """
        definition = loader.load_agent_yaml(file)
        return cls(
            name=definition.get('name'),
            in_selectors=definition.get('in_selectors'),
            out_selectors=definition.get('out_selectors'),
            args=definition.get('args'),
            constraints=definition.get('constraints'),
            mounts=definition.get('mounts'),
            restart_policy=definition.get('restart_policy'),
            mem_limit=definition.get('mem_limit'),
            open_ports=definition.get('open_ports'),
        )



"""Handles any configurations related to Ostorlab, such as storing and retriving
    tokens or API keys.

    This module contains code to handle any configurations related to Ostorlab
    such as storing and retriving API keys.
"""

from typing import Dict, Optional
from datetime import datetime
import json
import pathlib

OSTORLAB_PRIVATE_DIR = pathlib.Path.home() / '.ostorlab'


class ConfigurationManager:
    """Handles any configurations related to Ostorlab, such as storing and retriving
       API keys.
    """

    def __init__(self, private_dir: str = OSTORLAB_PRIVATE_DIR):
        """Constructs all the necessary attributes for the object

        Args:
            private_dir: The private directory where Ostorlab configurations are stored.
            Defaults to OSTORLAB_PRIVATE_DIR.
        """
        self._private_dir = private_dir
        self._private_dir.mkdir(parents=True, exist_ok=True)
        self._complete_api_key_path = self._private_dir / 'key'

    def set_api_data(self, secret_key: str, api_key_id: str, expiry_date: Optional[datetime]) -> None:
        """Persists the API data (key, id, expiry date) to a file in the given path.

        Args:
            api_key: The authenticated user's generated API key.
        """

        api_data = {
            'secret_key': secret_key,
            'api_key_id': api_key_id,
            'expiry_date': expiry_date
        }

        with open(self._complete_api_key_path, 'w', encoding='utf-8') as file:
            data = json.dumps(api_data, indent=4)
            file.write(data)

    def get_api_data(self) -> Optional[Dict]:
        """Gets the API data from the location in which it is saved.

        Returns:
            The user's API data if it exists, otherwise returns None.
        """
        try:
            with open(self._complete_api_key_path, 'r', encoding='utf-8') as file:
                return json.loads(file.read())
        except FileNotFoundError:
            return None

    def get_api_key(self) -> Optional[str]:
        """Gets the API key from the location in which it is saved.

        Returns:
            The user's API key if it exists, otherwise returns None.
        """
        api_data = self.get_api_data()
        if api_data is not None:
            return api_data.get('secret_key')
        else:
            return None

    def get_api_key_id(self) -> Optional[str]:
        """Gets the API key id from the location in which it is saved.

        Returns:
            The user's API key id if it exists, otherwise returns None.
        """
        api_data = self.get_api_data()
        if api_data is not None:
            return api_data.get('api_key_id')
        else:
            return None

    def get_api_key_expiry_date(self) -> Optional[datetime]:
        """Gets the API key expiry date from the location in which it is saved.

        Returns:
            The user's API key expiry date if it exists, otherwise returns None.
        """
        api_data = self.get_api_data()
        if api_data is not None:
            return api_data.get('expiry_date')
        else:
            return None


    def delete_api_data(self) -> None:
        """Deletes the file containing the API data.
        """
        self._complete_api_key_path.unlink(missing_ok=True)



"""Revokes the API key."""

from ostorlab.apis import request
from typing import Dict, Optional
import json
import logging

logger = logging.getLogger(__name__)


class RevokeAPIKeyAPIRequest(request.APIRequest):
    """Revokes the API key."""

    def __init__(self, api_key_id: str):
        """Sets API key identifier.

        Args:
           api_key_id: The API key id used to revoke the API key.
        """
        self._api_key_id = api_key_id

    @property
    def query(self) -> Optional[str]:
        """Defines the query to revoke the API key.

        Returns:
            The query to revoke the API key
        """
        return """
         mutation RevokeApiKey($apiKeyId: String!) {
               revokeApiKey(apiKeyId: $apiKeyId) {
                  result
               }
            }
        """

    @property
    def endpoint(self):
        return request.AUTHENTICATED_GRAPHQL_ENDPOINT

    @property
    def data(self) -> Optional[Dict]:
        """Sets the query and variables to revoke the API key.

        Returns:
              The query and variables to revoke the API key.
        """
        data = {
            'query': self.query,
            'variables': json.dumps({'apiKeyId': self._api_key_id})
        }
        return data



"""Local runtime runs agents locally.

The local runtime requires Docker Swarm to run robust long-running services with a set of configured services, like
a local RabbitMQ.
"""
import logging
import socket
from typing import List
from typing import Optional

import docker
import requests
import tenacity
from docker.models import services as docker_models_services
from docker.types import services as docker_types_services

from ostorlab import exceptions
from ostorlab.assets import asset as base_asset
from ostorlab.runtimes import definitions
from ostorlab.runtimes import runtime
from ostorlab.runtimes.local.services import mq
from ostorlab.utils import strings as strings_utils

NETWORK = 'ostorlab_local_network'
HEALTHCHECK_HOST = '0.0.0.0'
HEALTHCHECK_PORT = '5000'

logger = logging.getLogger(__name__)

ASSET_INJECTION_AGENT_DEFAULT = 'agent/ostor/inject_asset'


class UnhealthyService(exceptions.OstorlabError):
    """A service by the runtime is considered unhealthy."""


def _is_agent_status_ok(ip: str) -> bool:
    """Agent are expected to expose a healthcheck service on port 5000 that returns status code 200 and `OK` response.

    Args:
        ip: The agent IP address.

    Returns:
        True if the agent is healthy, false otherwise.
    """
    status_ok = False
    try:
        status_ok = requests.get(f'http://{ip}:5000/status').text == 'OK'
    except requests.exceptions.ConnectionError:
        logger.error('unable to connect to %s', ip)
    return status_ok


def _get_task_ips(service: docker_models_services.Service) -> List[str]:
    """Returns list of IP addresses assigned to the tasks of a docker service.

    Args:
        service: docker service.

    Returns:
        List of IP addresses.
    """
    # current implementation supports only one task per service.
    logger.info('getting ips for task %s', service.name)
    try:
        ips = socket.gethostbyname_ex(f'tasks.{service.name}')
        logger.info('found ips %s for task %s', ips, service.name)
        return ips[2]
    except socket.gaierror:
        return []


def _is_service_type_run(service: docker_models_services.Service) -> bool:
    """Checks if the service should run once or should be continuously running.

    Args:
        service: Docker service.

    Returns:
        Bool indicating if the service is run-once or long-running.
    """
    return service.attrs['Spec']['TaskTemplate']['RestartPolicy']['Condition'] == 'none'


class LocalRuntime(runtime.Runtime):
    """Local runtime runes agents locally using Docker Swarm.
    Local runtime starts a Vanilla RabbitMQ service, starts all the agents listed in the `AgentRunDefinition`, checks
    their status and then inject the target asset.
    """

    def __init__(self, name: Optional[str] = None, network: str = NETWORK) -> None:
        super().__init__()
        self._name = name or strings_utils.random_string(6)
        self._network = network
        self._mq_service: Optional[mq.LocalRabbitMQ] = None
        # TODO(alaeddine): inject docker client to support more complex use-cases.
        self._docker_client = docker.from_env()

    @property
    def name(self) -> str:
        """Local runtime instance name."""
        return self._name

    def can_run(self, agent_group_definition: definitions.AgentGroupDefinition) -> bool:
        """Checks if the runtime can run the provided agent run definition.

        Args:
            agent_group_definition: Agent and Agent group definition.

        Returns:
            Always true for the moment as the local runtime don't have restrictions on what it can run.
        """
        del agent_group_definition
        return True

    def scan(self, agent_group_definition: definitions.AgentGroupDefinition, asset: base_asset.Asset) -> None:
        """Start scan on asset using the provided agent run definition.

        The scan takes care of starting all the scan required services, ensuring they are healthy, starting all the
         agents, ensuring they are healthy and then injects the target asset.

        Args:
            agent_group_definition: Agent run definition defines the set of agents and how agents are configured.
            asset: the target asset to scan.

        Returns:
            None
        """
        self._create_network()
        self._start_services()
        self._check_services_healthy()
        self._start_agents(agent_group_definition)
        self._check_agents_healthy(agent_group_definition)
        self._inject_asset(asset)

    def _create_network(self):
        """Creates a docker swarm network where all services and agents can communicates."""
        if any(network.name == self._network for network in self._docker_client.networks.list()):
            logger.warning('network already exists.')
        else:
            logger.info('creating private network %s', self._network)
            return self._docker_client.networks.create(
                name=self._network,
                driver='overlay',
                attachable=True,
                labels={'ostorlab.universe': self._name},
                check_duplicate=True
            )

    def _start_services(self):
        """Start all the local runtime services."""
        self._start_mq_service()

    def _start_mq_service(self):
        """Start a local rabbitmq service."""
        self._mq_service = mq.LocalRabbitMQ(name=self._name, network=self._network)
        self._mq_service.start()

    def _check_services_healthy(self):
        """Check if the rabbitMQ service is running and healthy."""
        if self._mq_service is None or self._mq_service.is_healthy is False:
            raise UnhealthyService('MQ service is unhealthy.')

    def _start_agents(self, agent_group_definition: definitions.AgentGroupDefinition):
        """Starts all the agents as list in the agent run definition."""
        for agent in agent_group_definition.agents:
            agent_instance_settings_proto = agent.to_raw_proto()
            docker_config = self._docker_client.configs.create(name=f'agent_{agent.key}_{self._name}',
                                                               data=agent_instance_settings_proto)
            config_reference = docker.types.ConfigReference(config_id=docker_config.id,
                                                              config_name=f'agent_{agent.key}_{self._name}',
                                                              filename='/tmp/settings.binproto')
            self._start_agent(agent, [config_reference])

    def _start_agent(self, agent: definitions.AgentSettings,
                     extra_configs: Optional[List[docker.types.ConfigReference]] = None) -> None:
        """Start agent based on provided definition.

        Args:
            agent: An agent definition containing all the settings of how agent should run and what arguments to pass.
        """
        logger.info('starting agent %s with %s', agent.key, agent.args)
        # Port published with ingress mode can't be used with dnsrr mode
        if agent.open_ports:
            endpoint_spec = docker_types_services.EndpointSpec(
                ports={p.destination_port: p.source_port for p in agent.open_ports})
        else:
            endpoint_spec = docker_types_services.EndpointSpec(mode='dnsrr')

        agent.bus_url = self._mq_service.url
        agent.bus_exchange_topic = f'ostorlab_topic_{self._name}'
        agent.healthcheck_host = HEALTHCHECK_HOST
        agent.healthcheck_port = HEALTHCHECK_PORT

        agent_service = self._docker_client.services.create(
            image=agent.container_image,
            networks=[self._network],
            env=[
                f'UNIVERSE={self._name}',
            ],
            name=f'{agent.container_image}_{self._name}',
            restart_policy=docker_types_services.RestartPolicy(condition=agent.restart_policy),
            mounts=agent.mounts,
            configs=extra_configs,
            labels={'ostorlab.universe': self._name},
            constraints=agent.constraints,
            endpoint_spec=endpoint_spec,
            resources=docker_types_services.Resources(mem_limit=agent.mem_limit))
        if agent.replicas > 1:
            # Ensure the agent service had to
            # TODO(alaeddine): Check if sleep if really needed and if it is, implement a parallel way to start agents
            #  and scale them.
            # time.sleep(10)
            self._scale_service(agent_service, agent.replicas)

    def _check_agents_healthy(self, agent_group_definition: definitions.AgentGroupDefinition):
        """Checks if an agent is healthy."""
        return self._are_agents_ready(agent_group_definition.agents)

    @tenacity.retry(stop=tenacity.stop_after_attempt(20),
                    wait=tenacity.wait_exponential(multiplier=1, max=12),
                    # return last value and don't raise RetryError exception.
                    retry_error_callback=lambda lv: lv.result(),
                    retry=tenacity.retry_if_result(lambda v: v is False))
    def _is_service_healthy(self, service: docker_models_services.Service, replicas=None) -> bool:
        """Checks if a docker service is healthy by checking all tasks status."""
        logger.info('checking service %s', service.name)
        if not replicas:
            replicas = service.attrs['Spec']['Mode']['Replicated']['Replicas']
        return replicas == len([task for task in service.tasks() if task['Status']['State'] == 'running'])

    @tenacity.retry(stop=tenacity.stop_after_attempt(20),
                    wait=tenacity.wait_exponential(multiplier=1, max=20),
                    # return last value and don't raise RetryError exception.
                    retry_error_callback=lambda lv: lv.result(),
                    retry=tenacity.retry_if_result(lambda v: v is False))
    def _are_agents_ready(self, agents: List[definitions.AgentSettings], fail_fast=True) -> bool:
        """Checks that all agents are ready and healthy while taking into account the run type of agent
         (once vs long-running)."""
        all_agents_healthy = True
        logger.info('listing services ...')
        agent_services = list(self._list_agent_services())
        if len(agent_services) < len(agents):
            logger.error('found %d, expecting %d', len(agent_services), len(agents))
            return False
        else:
            logger.info('found correct count of services')

        for service in agent_services:
            logger.info('checking %s ...', service.name)
            if not _is_service_type_run(service):
                if self._is_service_healthy(service):
                    task_healthy = False
                    container_ips = _get_task_ips(service)
                    healthy = any(_is_agent_status_ok(ip) for ip in container_ips)
                    # only a single task need to be healthy
                    task_healthy = task_healthy or healthy

                    # all agents need to be healthy
                    all_agents_healthy = all_agents_healthy and task_healthy
                    if fail_fast and not all_agents_healthy:
                        logger.error('agent health check %s is not healthy', service.name)
                        return False

                    logger.info('agent healthcheck of %s is %s', service.name, all_agents_healthy)
                else:
                    logger.error('agent service %s is not healthy', service.name)
                    if fail_fast and not all_agents_healthy:
                        return False
        return all_agents_healthy

    def _list_agent_services(self):
        """List the services of type agents. All agent service must start with agent_."""
        services = self._docker_client.services.list(filters={'label': f'ostorlab.universe={self._name}'})
        for service in services:
            if service.name.startswith('agent_'):
                yield service

    def _inject_asset(self, asset: base_asset.Asset):
        """Injects the scan target assets."""
        docker_config = self._docker_client.configs.create(name='asset',
                                                           data=asset.to_proto())
        config_reference = docker.types.ConfigReference(config_id=docker_config.id,
                                                        config_name='asset',
                                                        filename='/tmp/asset.binproto')
        inject_asset_agent_settings = definitions.AgentSettings(key=ASSET_INJECTION_AGENT_DEFAULT)
        self._start_agent(agent=inject_asset_agent_settings, extra_configs=[config_reference])

    def _scale_service(self, service: docker_models_services.Service, replicas: int) -> None:
        """Calling scale directly on the service causes an API error. This is a workaround that simulates refreshing
         the service object, then calling the scale API."""
        for s in self._docker_client.services.list():
            if s.name == service.name:
                s.scale(replicas)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: java_method_call.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='java_method_call.proto',
    package='',
    serialized_pb=_b(
        '\n\x16java_method_call.proto\"\x8a\x02\n\x10java_method_call\x12\x0f\n\x07scan_id\x18\x01 \x02(\x05\x12\x12\n\nclass_name\x18\x02 \x02(\t\x12\x13\n\x0bmethod_name\x18\x03 \x02(\t\x12\x11\n\tsignature\x18\x04 \x02(\t\x12/\n\nparameters\x18\x05 \x03(\x0b\x32\x1b.java_method_call.parameter\x1a/\n\x0bjava_object\x12\x11\n\tsignature\x18\x04 \x02(\t\x12\r\n\x05value\x18\x05 \x02(\x0c\x1aG\n\tparameter\x12\x0c\n\x04name\x18\x02 \x02(\t\x12,\n\x05value\x18\x03 \x02(\x0b\x32\x1d.java_method_call.java_object')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_JAVA_METHOD_CALL_JAVA_OBJECT = _descriptor.Descriptor(
    name='java_object',
    full_name='java_method_call.java_object',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='signature', full_name='java_method_call.java_object.signature', index=0,
            number=4, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='value', full_name='java_method_call.java_object.value', index=1,
            number=5, type=12, cpp_type=9, label=2,
            has_default_value=False, default_value=_b(""),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=173,
    serialized_end=220,
)

_JAVA_METHOD_CALL_PARAMETER = _descriptor.Descriptor(
    name='parameter',
    full_name='java_method_call.parameter',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='name', full_name='java_method_call.parameter.name', index=0,
            number=2, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='value', full_name='java_method_call.parameter.value', index=1,
            number=3, type=11, cpp_type=10, label=2,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=222,
    serialized_end=293,
)

_JAVA_METHOD_CALL = _descriptor.Descriptor(
    name='java_method_call',
    full_name='java_method_call',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='scan_id', full_name='java_method_call.scan_id', index=0,
            number=1, type=5, cpp_type=1, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='class_name', full_name='java_method_call.class_name', index=1,
            number=2, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='method_name', full_name='java_method_call.method_name', index=2,
            number=3, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='signature', full_name='java_method_call.signature', index=3,
            number=4, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='parameters', full_name='java_method_call.parameters', index=4,
            number=5, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[_JAVA_METHOD_CALL_JAVA_OBJECT, _JAVA_METHOD_CALL_PARAMETER, ],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=27,
    serialized_end=293,
)

_JAVA_METHOD_CALL_JAVA_OBJECT.containing_type = _JAVA_METHOD_CALL
_JAVA_METHOD_CALL_PARAMETER.fields_by_name['value'].message_type = _JAVA_METHOD_CALL_JAVA_OBJECT
_JAVA_METHOD_CALL_PARAMETER.containing_type = _JAVA_METHOD_CALL
_JAVA_METHOD_CALL.fields_by_name['parameters'].message_type = _JAVA_METHOD_CALL_PARAMETER
DESCRIPTOR.message_types_by_name['java_method_call'] = _JAVA_METHOD_CALL

java_method_call = _reflection.GeneratedProtocolMessageType('java_method_call', (_message.Message,), dict(

    java_object=_reflection.GeneratedProtocolMessageType('java_object', (_message.Message,), dict(
        DESCRIPTOR=_JAVA_METHOD_CALL_JAVA_OBJECT,
        __module__='java_method_call_pb2'
        # @@protoc_insertion_point(class_scope:java_method_call.java_object)
    ))
    ,

    parameter=_reflection.GeneratedProtocolMessageType('parameter', (_message.Message,), dict(
        DESCRIPTOR=_JAVA_METHOD_CALL_PARAMETER,
        __module__='java_method_call_pb2'
        # @@protoc_insertion_point(class_scope:java_method_call.parameter)
    ))
    ,
    DESCRIPTOR=_JAVA_METHOD_CALL,
    __module__='java_method_call_pb2'
    # @@protoc_insertion_point(class_scope:java_method_call)
))
_sym_db.RegisterMessage(java_method_call)
_sym_db.RegisterMessage(java_method_call.java_object)
_sym_db.RegisterMessage(java_method_call.parameter)

# @@protoc_insertion_point(module_scope)
